"""Serializer for Subscriptions."""

from typing import Any

from rest_framework import serializers


class SubscriptionSerializer(serializers.Serializer[dict[str, Any]]):
    """
    Serializer for Subscription data.

    This serializer is used to serialize and deserialize subscription data.
    """

    id = serializers.CharField(source="_id")
    subscriber_id = serializers.CharField()
    source_id = serializers.CharField()
    source_type = serializers.CharField()

    def create(self, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError

    def update(self, instance: Any, validated_data: dict[str, Any]) -> Any:
        """Raise NotImplementedError"""
        raise NotImplementedError



"""Subscriptions API Views."""

from rest_framework import status
from rest_framework.permissions import AllowAny
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

from forum.api.subscriptions import (
    create_subscription,
    delete_subscription,
    get_thread_subscriptions,
    get_user_subscriptions,
)
from forum.pagination import ForumPagination
from forum.utils import ForumV2RequestError


class SubscriptionAPIView(APIView):
    """
    API View for managing subscriptions.

    This view provides endpoints for subscribing and unsubscribing users to/from sources.
    """

    permission_classes = (AllowAny,)

    def post(self, request: Request, user_id: str) -> Response:
        """
        Subscribe a user to a source.

        Args:
            request (HttpRequest): The HTTP request object.
            user_id (str): The ID of the user to subscribe.

        Returns:
            Response: A Response object with the subscription data.

        Raises:
            HTTP_400_BAD_REQUEST: If the user or content does not exist.
        """
        request_data = request.data
        try:
            serilized_data = create_subscription(user_id, request_data["source_id"])
        except ForumV2RequestError as e:
            return Response(data={"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)
        return Response(data=serilized_data, status=status.HTTP_200_OK)

    def delete(self, request: Request, user_id: str) -> Response:
        """
        Unsubscribe a user from a source.

        Args:
            request (HttpRequest): The HTTP request object.
            user_id (str): The ID of the user to unsubscribe.

        Returns:
            Response: A Response object with the subscription data.

        Raises:
            HTTP_400_BAD_REQUEST: If the user or subscription does not exist.
        """
        try:
            params = request.query_params
            serilized_data = delete_subscription(user_id, params["source_id"])
        except ForumV2RequestError as e:
            return Response(data={"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)
        return Response(data=serilized_data, status=status.HTTP_200_OK)


class UserSubscriptionAPIView(APIView):
    """
    API View for managing user subscriptions.

    This view provides an endpoint for retrieving a user's subscriptions.
    """

    permission_classes = (AllowAny,)

    def get(self, request: Request, user_id: str) -> Response:
        """
        Retrieve a user's subscriptions.

        Args:
            request (HttpRequest): The HTTP request object.
            user_id (str): The ID of the user to retrieve subscriptions for.

        Returns:
            Response: A Response object with the subscription data.

        Raises:
            HTTP_400_BAD_REQUEST: If the user does not exist.
        """
        params = request.GET.dict()
        try:
            serilized_data = get_user_subscriptions(
                user_id, params["course_id"], params
            )
        except ForumV2RequestError as e:
            return Response(data={"error": str(e)}, status=status.HTTP_400_BAD_REQUEST)
        return Response(data=serilized_data, status=status.HTTP_200_OK)


class ThreadSubscriptionAPIView(APIView):
    """
    API View for managing thread subscriptions.

    This view provides an endpoint for retrieving subscriptions to a specific thread.
    """

    permission_classes = (AllowAny,)
    pagination_class = ForumPagination

    def get(self, request: Request, thread_id: str) -> Response:
        """
        Retrieve subscriptions to a specific thread.

        Args:
            request (HttpRequest): The HTTP request object.
            thread_id (str): The ID of the thread to retrieve subscriptions for.

        Returns:
            Response: A paginated Response object with the subscription data.
        """
        page = int(request.GET.get("page", 1))
        per_page = int(request.GET.get("per_page", 20))
        subscriptions_data = get_thread_subscriptions(thread_id, page, per_page)
        return Response(subscriptions_data, status=status.HTTP_200_OK)



"""Management command for initialize forum indices"""

from argparse import ArgumentParser
from typing import Any

from django.core.management.base import BaseCommand

from forum.search import get_index_search_backend


class Command(BaseCommand):
    """
    Django management command for the initialization of search indices.
    """

    help = "Initialize Elasticsearch indices, optionally forcing the creation of new indices."

    def add_arguments(self, parser: ArgumentParser) -> None:
        """
        Adds command line arguments to the initialize_forum_indices command.

        Args:
            parser: The argument parser to which the --force argument is added.
        """
        parser.add_argument(
            "--force",
            action="store_true",
            help="Force the creation of new indices even if they exist.",
        )

    def handle(self, *args: list[str], **kwargs: dict[str, Any]) -> None:
        """
        Handles the execution of the initialize_indices command.

        Initializes Elasticsearch indices. If the --force option is provided,
        it forces the creation of new indices even if they exist.

        Args:
            args: Additional arguments.
            kwargs: Command options.
        """
        search_backend = get_index_search_backend()
        force_new_index = bool(kwargs.get("force", False))
        search_backend.initialize_indices(force_new_index=force_new_index)
        self.stdout.write(self.style.SUCCESS("Forum indices initialized successfully."))



"""
These settings are here to use during tests, because django requires them.

In a real-world use case, apps in this project are installed into other
Django applications, so these settings will not be used.
"""

from os.path import abspath, dirname, join


def root(*args: str) -> str:
    """
    Get the absolute path of the given path relative to the project root.
    """
    return join(abspath(dirname(__file__)), *args)


DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "default.db",
        "USER": "",
        "PASSWORD": "",
        "HOST": "",
        "PORT": "",
    }
}

INSTALLED_APPS = (
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.messages",
    "django.contrib.sessions",
    "forum",
)

LOCALE_PATHS = [
    root("forum", "conf", "locale"),
]

ROOT_URLCONF = "forum.urls"

SECRET_KEY = "insecure-secret-key"

MIDDLEWARE = (
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
)

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": False,
        "OPTIONS": {
            "context_processors": [
                "django.contrib.auth.context_processors.auth",  # this is required for admin
                "django.contrib.messages.context_processors.messages",  # this is required for admin
            ],
        },
    }
]

FORUM_MONGODB_DATABASE = "testdb"
FORUM_MONGODB_CLIENT_PARAMETERS: dict[str, str] = {}

FORUM_SEARCH_BACKEND = "forum.search.es.ElasticsearchBackend"
FORUM_ELASTIC_SEARCH_CONFIG = [
    {
        "host": "localhost",
        "port": "5200",
        "use_ssl": False,
    }
]

# Meilisearch connection parameters
MEILISEARCH_URL = "http://localhost:5700"
MEILISEARCH_API_KEY = "MEILISEARCH_MASTER_KEY"

USE_TZ = True



"""Forum v2 feature toggles."""

# pylint: disable=E0401
from openedx.core.djangoapps.waffle_utils import CourseWaffleFlag  # type: ignore[import-not-found]

FORUM_V2_WAFFLE_FLAG_NAMESPACE = "forum_v2"


# .. toggle_name: forum_v2.enable_mysql_backend
# .. toggle_implementation: CourseWaffleFlag
# .. toggle_default: False
# .. toggle_description: Waffle flag to use the MySQL backend instead of Mongo backend.
# .. toggle_use_cases: temporary, open_edx
# .. toggle_creation_date: 2024-10-18
# .. toggle_target_removal_date: 2025-06-18
ENABLE_MYSQL_BACKEND = CourseWaffleFlag(
    f"{FORUM_V2_WAFFLE_FLAG_NAMESPACE}.enable_mysql_backend", __name__
)



"""Content Class for mongo backend."""

from datetime import datetime
from typing import Any, Optional

from bson import ObjectId
from pymongo.cursor import Cursor

from forum.backends.mongodb.base_model import MongoBaseModel


class BaseContents(MongoBaseModel):
    """
    BaseContents class: same as Contents, but without "update" and "insert" methods,
    because child classes will have different signatures for these methods.
    """

    content_type: str = ""
    COLLECTION_NAME: str = "contents"

    def create_indexes(self) -> None:
        """
        The implementation creates the indexes in the mongodb for the contents collection.
        """
        self._collection.create_index(
            [
                ("_type", 1),
                ("course_id", 1),
                ("pinned", -1),
                ("created_at", -1),
            ],
            background=True,
        )
        self._collection.create_index(
            [
                ("_type", 1),
                ("course_id", 1),
                ("pinned", -1),
                ("comment_count", -1),
                ("created_at", -1),
            ],
            background=True,
        )
        self._collection.create_index(
            [
                ("_type", 1),
                ("course_id", 1),
                ("pinned", -1),
                ("votes.point", -1),
                ("created_at", -1),
            ],
            background=True,
        )
        self._collection.create_index(
            [
                ("_type", 1),
                ("course_id", 1),
                ("pinned", -1),
                ("last_activity_at", -1),
                ("created_at", -1),
            ],
            background=True,
        )
        self._collection.create_index(
            [
                ("comment_thread_id", 1),
                ("sk", 1),
            ],
            sparse=True,
        )
        self._collection.create_index(
            [
                ("comment_thread_id", 1),
                ("endorsed", 1),
            ],
            sparse=True,
        )
        self._collection.create_index(
            [
                ("commentable_id", 1),
            ],
            sparse=True,
            background=True,
        )

    @classmethod
    def mapping(cls) -> dict[str, Any]:
        """
        Implement this method in the child class
        """
        raise NotImplementedError

    @classmethod
    def doc_to_hash(cls, doc: dict[str, Any]) -> dict[str, Any]:
        """
        Implement this method in the child class
        """
        raise NotImplementedError

    def override_query(self, query: dict[str, Any]) -> dict[str, Any]:
        """
        Override the query with the _type field.
        """
        query = {**query, "_type": self.content_type}
        return super().override_query(query)

    def get_list(self, **kwargs: Any) -> Cursor[dict[str, Any]]:
        """
        Retrieves a list of all content documents in the database based on provided filters.

        Args:
            kwargs: The filter arguments.

        Returns:
            A list of content documents.
        """
        resp_skip = kwargs.pop("resp_skip", 0)
        resp_limit = kwargs.pop("resp_limit", None)

        if self.content_type:
            kwargs["_type"] = self.content_type

        # Apply sorting first if provided
        sort = kwargs.pop("sort", None)
        query = self._collection.find(kwargs)
        if sort:
            query = query.sort("sk", sort)

        # Apply pagination after sorting
        query = query.skip(resp_skip)
        if resp_limit is not None:
            query = query.limit(resp_limit)

        return query

    @classmethod
    def get_votes_dict(cls, up: list[str], down: list[str]) -> dict[str, Any]:
        """
        Calculates and returns the vote summary for a thread.

        Args:
            up (list): A list of user IDs who upvoted the thread.
            down (list): A list of user IDs who downvoted the thread.

        Returns:
            dict: A dictionary containing the vote summary with the following keys:
                - "up" (list): The list of user IDs who upvoted.
                - "down" (list): The list of user IDs who downvoted.
                - "up_count" (int): The count of upvotes.
                - "down_count" (int): The count of downvotes.
                - "count" (int): The total number of votes (upvotes + downvotes).
                - "point" (int): The vote score (upvotes - downvotes).
        """
        up = up or []
        down = down or []
        votes = {
            "up": up,
            "down": down,
            "up_count": len(up),
            "down_count": len(down),
            "count": len(up) + len(down),
            "point": len(up) - len(down),
        }
        return votes

    def update_votes(self, content_id: str, votes: dict[str, Any]) -> int:
        """
        Updates a votes in the content document.

        Args:
        content_id: The id of the content model
        votes (Optional[dict[str, int]], optional): The votes for the thread.
        """
        update_data = {"votes": votes, "updated_at": datetime.now()}
        result = self._collection.update_one(
            {"_id": ObjectId(content_id)},
            {"$set": update_data},
        )
        return result.modified_count

    def update_count(self, content_id: str, query: dict[str, Any]) -> int:
        """
        Updates count of a field in the content document based on query.

        Args:
            content_id (str): The id of the content(Commentthread id or Comment id) model.
            query (dict[str, Any]): Query to update the count in a specific field.
        """
        result = self._collection.update_one(
            {"_id": ObjectId(content_id)},
            query,
        )
        return result.modified_count


class Contents(BaseContents):
    """
    Contents class for cs_comments_service contents collection
    """

    def insert(
        self,
        _id: str,
        author_id: str,
        abuse_flaggers: list[str],
        historical_abuse_flaggers: list[str],
        visible: bool,
    ) -> str:
        """
        Inserts a new content document into the database.

        Args:
            _id (str): The ID of the content.
            author_id (str): The ID of the author who created the content.
            abuse_flaggers (list[str]): A list of IDs of users who flagged the content as abusive.
            historical_abuse_flaggers (list[str]): A list of IDs of users who previously flagged the content as abusive.
            visible (bool): Whether the content is visible or not.

        Returns:
            str: The ID of the inserted document.
        """
        content_data = {
            "_id": ObjectId(_id),
            "author_id": author_id,
            "abuse_flaggers": abuse_flaggers,
            "historical_abuse_flaggers": historical_abuse_flaggers,
            "visible": visible,
        }
        result = self._collection.insert_one(content_data)
        return str(result.inserted_id)

    def update(
        self,
        _id: str,
        author_username: Optional[str] = None,
        abuse_flaggers: Optional[list[str]] = None,
        historical_abuse_flaggers: Optional[list[str]] = None,
        body: Optional[str] = None,
        title: Optional[str] = None,
        **kwargs: Any
    ) -> int:
        """
        Updates a contents document in the database based on the provided _id.

        Args:
            _id: The id of the contents document to update.
            **kwargs: The fields to update in the contents document.

        Returns:
            The number of documents modified.
        """
        fields = [
            ("author_username", author_username),
            ("abuse_flaggers", abuse_flaggers),
            ("historical_abuse_flaggers", historical_abuse_flaggers),
            ("body", body),
            ("title", title),
        ]
        update_data: dict[str, Any] = {
            field: value for field, value in fields if value is not None
        }

        result = self._collection.update_one(
            {"_id": ObjectId(_id)},
            {"$set": update_data},
        )
        return result.modified_count



"""
Native Python Pins APIs.
"""

import logging
from typing import Any, Optional

from forum.backend import get_backend
from forum.serializers.thread import ThreadSerializer
from forum.utils import ForumV2RequestError

log = logging.getLogger(__name__)


def pin_unpin_thread(
    user_id: str,
    thread_id: str,
    action: str,
    course_id: Optional[str] = None,
) -> dict[str, Any]:
    """
    Helper method to Pin or Unpin a thread.
    Parameters:
        user_id (str): The ID of the requested User.
        thread_id (str): The ID of the thread to pin.
        action: (str): It's value can be "pin" or "unpin".
    Response:
        A response with the updated thread data.
    """
    backend = get_backend(course_id)()
    try:
        thread_data: dict[str, Any] = backend.handle_pin_unpin_thread_request(
            user_id, thread_id, action, ThreadSerializer
        )
    except ValueError as e:
        log.error(f"Forumv2RequestError for {action} thread request.")
        raise ForumV2RequestError(str(e)) from e

    return thread_data


def pin_thread(
    user_id: str, thread_id: str, course_id: Optional[str] = None
) -> dict[str, Any]:
    """
    Pin a thread.
    Parameters:
        user_id (str): The ID of the requested User.
        thread_id (str): The ID of the thread to pin.
    Response:
        A response with the updated thread data.
    """

    return pin_unpin_thread(user_id, thread_id, "pin", course_id)


def unpin_thread(
    user_id: str, thread_id: str, course_id: Optional[str] = None
) -> dict[str, Any]:
    """
    Unpin a thread.
    Parameters:
        user_id (str): The ID of the requested User.
        thread_id (str): The ID of the thread to pin.
    Response:
        A response with the updated thread data.
    """
    return pin_unpin_thread(user_id, thread_id, "unpin", course_id)



"""
URLs for forum.
"""

from django.urls import include, path

from forum.views.commentables import CommentablesCountAPIView
from forum.views.comments import CommentsAPIView, CreateThreadCommentAPIView
from forum.views.flags import CommentFlagAPIView, ThreadFlagAPIView
from forum.views.pins import PinThreadAPIView, UnpinThreadAPIView
from forum.views.search import SearchThreadsView
from forum.views.subscriptions import (
    SubscriptionAPIView,
    ThreadSubscriptionAPIView,
    UserSubscriptionAPIView,
)
from forum.views.threads import CreateThreadAPIView, ThreadsAPIView, UserThreadsAPIView
from forum.views.users import (
    UserActiveThreadsAPIView,
    UserAPIView,
    UserCourseStatsAPIView,
    UserCreateAPIView,
    UserEditAPIView,
    UserReadAPIView,
    UserRetireAPIView,
)
from forum.views.votes import CommentVoteView, ThreadVoteView

api_patterns = [
    # thread votes APIs
    path(
        "threads/<str:thread_id>/votes",
        ThreadVoteView.as_view(),
        name="thread-vote",
    ),
    path(
        "comments/<str:comment_id>/votes",
        CommentVoteView.as_view(),
        name="comment-vote",
    ),
    # abuse comment/thread APIs
    path(
        "comments/<str:comment_id>/abuse_<str:action>",
        CommentFlagAPIView.as_view(),
        name="comment-flags-api",
    ),
    path(
        "threads/<str:thread_id>/abuse_<str:action>",
        ThreadFlagAPIView.as_view(),
        name="thread-flags-api",
    ),
    # pin/unpin thread APIs
    path("threads/<str:thread_id>/pin", PinThreadAPIView.as_view(), name="pin-thread"),
    path(
        "threads/<str:thread_id>/unpin",
        UnpinThreadAPIView.as_view(),
        name="unpin-thread",
    ),
    # comments API
    path(
        "comments/<str:comment_id>",
        CommentsAPIView.as_view(),
        name="comments-api",
    ),
    path(
        "threads/<str:thread_id>/comments",
        CreateThreadCommentAPIView.as_view(),
        name="create-parent-comment-api",
    ),
    # search threads API
    path(
        "search/threads",
        SearchThreadsView.as_view(),
        name="search-thread-api",
    ),
    # subscription APIs
    path(
        "users/<str:user_id>/subscriptions",
        SubscriptionAPIView.as_view(),
        name="subscriptions",
    ),
    path(
        "users/<str:user_id>/subscribed_threads",
        UserSubscriptionAPIView.as_view(),
        name="user-subscriptions",
    ),
    path(
        "threads/<str:thread_id>/subscriptions",
        ThreadSubscriptionAPIView.as_view(),
        name="thread-subscriptions",
    ),
    # threads API
    path(
        "course/threads",
        CreateThreadAPIView.as_view(),
        name="create-thread-api",
    ),
    path(
        "threads",
        UserThreadsAPIView.as_view(),
        name="user-threads-api",
    ),
    path(
        "threads/<str:thread_id>",
        ThreadsAPIView.as_view(),
        name="threads-api",
    ),
    # commentables API
    path(
        "commentables/<str:course_id>/counts",
        CommentablesCountAPIView.as_view(),
        name="commentables-count",
    ),
    path(
        "users",
        UserCreateAPIView.as_view(),
        name="create-user",
    ),
    path(
        "users/<str:user_id>",
        UserAPIView.as_view(),
        name="get-user-detail",
    ),
    path(
        "users/<str:user_id>/replace_username",
        UserEditAPIView.as_view(),
        name="edit-user",
    ),
    path(
        "users/<str:user_id>/read",
        UserReadAPIView.as_view(),
        name="user-read",
    ),
    path(
        "users/<str:user_id>/active_threads",
        UserActiveThreadsAPIView.as_view(),
        name="user-active-threads",
    ),
    path(
        "users/<str:course_id>/stats",
        UserCourseStatsAPIView.as_view(),
        name="user-course-stats",
    ),
    path(
        "users/<str:course_id>/update_stats",
        UserCourseStatsAPIView.as_view(),
        name="user-course-stats-update",
    ),
    path(
        "users/<str:user_id>/retire",
        UserRetireAPIView.as_view(),
        name="user-retire",
    ),
    # Proxy view for various API endpoints
    # Uncomment to redirect remaining API calls to the V1 API.
    # path(
    #     "<path:suffix>",
    #     ForumProxyAPIView.as_view(),
    #     name="forum_proxy",
    # ),
]

urlpatterns = [
    # for backward compatibility with edx-platform
    path("api/v1/", include(api_patterns)),
    # for when we will migrate edx-platform to the v2/ prefix
    path("api/v2/", include(api_patterns)),
]



"""
Native Python APIs.
"""

from .commentables import get_commentables_stats
from .comments import (
    create_child_comment,
    create_parent_comment,
    delete_comment,
    get_course_id_by_comment,
    get_parent_comment,
    update_comment,
)
from .flags import (
    update_comment_flag,
    update_thread_flag,
)
from .pins import pin_thread, unpin_thread
from .search import search_threads
from .subscriptions import (
    create_subscription,
    delete_subscription,
    get_thread_subscriptions,
    get_user_subscriptions,
)
from .threads import (
    create_thread,
    delete_thread,
    get_course_id_by_thread,
    get_thread,
    get_user_threads,
    update_thread,
)
from .users import (
    create_user,
    get_user,
    get_user_active_threads,
    get_user_course_stats,
    mark_thread_as_read,
    retire_user,
    update_user,
    update_username,
    update_users_in_course,
)
from .votes import (
    delete_comment_vote,
    delete_thread_vote,
    update_comment_votes,
    update_thread_votes,
)

__all__ = [
    "create_child_comment",
    "create_parent_comment",
    "create_subscription",
    "create_thread",
    "create_user",
    "delete_comment",
    "delete_comment_vote",
    "delete_subscription",
    "delete_thread",
    "delete_thread_vote",
    "get_commentables_stats",
    "get_course_id_by_comment",
    "get_course_id_by_thread",
    "get_parent_comment",
    "get_thread",
    "get_thread_subscriptions",
    "get_user",
    "get_user_active_threads",
    "get_user_course_stats",
    "get_user_subscriptions",
    "get_user_threads",
    "mark_thread_as_read",
    "pin_thread",
    "retire_user",
    "search_threads",
    "unpin_thread",
    "update_comment",
    "update_comment_flag",
    "update_comment_votes",
    "update_thread",
    "update_thread_flag",
    "update_thread_votes",
    "update_user",
    "update_username",
    "update_users_in_course",
]



"""
forum Django application initialization.
"""

from django.apps import AppConfig


class ForumConfig(AppConfig):
    """
    Configuration for the forum Django application.
    """

    name = "forum"

    plugin_app = {
        "url_config": {
            "lms.djangoapp": {
                "namespace": "forum",
                "regex": r"^forum/",
                "relative_path": "urls",
            }
        },
        "settings_config": {
            "lms.djangoapp": {
                "common": {"relative_path": "settings.common"},
                "devstack": {"relative_path": "settings.devstack"},
                "production": {"relative_path": "settings.production"},
            }
        },
    }

    def ready(self) -> None:
        """
        Import Signals.
        """
        import forum.signals  # pylint: disable=import-outside-toplevel, unused-import



"""
Signals for the forum App.
"""

from django.dispatch import Signal

from forum import handlers

# Those objects are useful and should be preserved for testing purposes.
comment_deleted = Signal()
comment_thread_deleted = Signal()
comment_inserted = Signal()
comment_thread_inserted = Signal()
comment_updated = Signal()
comment_thread_updated = Signal()

# Connect the handlers with the search backend
comment_deleted.connect(handlers.handle_comment_deletion)
comment_thread_deleted.connect(handlers.handle_comment_thread_deletion)
comment_inserted.connect(handlers.handle_comment_insertion)
comment_thread_inserted.connect(handlers.handle_comment_thread_insertion)
comment_updated.connect(handlers.handle_comment_updated)
comment_thread_updated.connect(handlers.handle_comment_thread_updated)



"""Pagination class for forum api."""

from typing import Any

from rest_framework.exceptions import NotFound
from rest_framework.pagination import PageNumberPagination
from rest_framework.request import Request


class ForumPagination(PageNumberPagination):
    """
    Custom pagination class for thread subscriptions.

    This class allows for custom pagination settings.
    """

    page_size_query_param = "per_page"
    max_page_size = 20

    def get_page_size(self, request: Request) -> int:
        """
        Get the page size from the request's GET parameters.

        If the `per_page` parameter is not provided, use the default page size from settings.
        """
        page_size = request.GET.get("per_page")
        if page_size is None:
            return self.max_page_size
        return min(int(page_size), self.max_page_size)

    def paginate_queryset(
        self, queryset: Any, request: Request, view: Any = None
    ) -> Any:
        """
        Paginate the queryset.

        If the page is invalid, return a custom error response instead of raising a 404 error.
        """
        try:
            return super().paginate_queryset(queryset, request, view)
        except NotFound:
            return []



"""
Common settings for forum app.
"""

from typing import Any


def plugin_settings(settings: Any) -> None:
    """
    Common settings for forum app
    """
    # Search backend
    if getattr(settings, "MEILISEARCH_ENABLED", False):
        settings.FORUM_SEARCH_BACKEND = getattr(
            settings,
            "FORUM_SEARCH_BACKEND",
            "forum.search.meilisearch.MeilisearchBackend",
        )
    else:
        settings.FORUM_SEARCH_BACKEND = getattr(
            settings, "FORUM_SEARCH_BACKEND", "forum.search.es.ElasticsearchBackend"
        )
        settings.FORUM_ELASTIC_SEARCH_CONFIG = getattr(
            settings, "FORUM_ELASTIC_SEARCH_CONFIG", [{"host": "elasticsearch"}]
        )

    # Unfortunately we can't copy settings from edx-platform because tutor patches have
    # not been applied yet
    settings.FORUM_MONGODB_DATABASE = getattr(
        settings, "FORUM_MONGODB_DATABASE", "cs_comments_service"
    )
    settings.FORUM_MONGODB_CLIENT_PARAMETERS = getattr(
        settings, "FORUM_MONGODB_CLIENT_PARAMETERS", {"host": "mongodb"}
    )

    # Enable forum service
    if "ENABLE_DISCUSSION_SERVICE" not in settings.FEATURES:
        settings.FEATURES["ENABLE_DISCUSSION_SERVICE"] = True

    # URL prefix must match the regex in the url_config of the plugin app
    settings.COMMENTS_SERVICE_URL = getattr(
        settings, "COMMENTS_SERVICE_URL", "http://localhost:8000/forum"
    )

    # Timezone-awareness is required for mysql fields
    settings.USE_TZ = getattr(settings, "USE_TZ", True)



"""
Mongo Models
"""

from .comments import Comment
from .contents import BaseContents, Contents
from .subscriptions import Subscriptions
from .threads import CommentThread
from .users import Users

__all__ = [
    "BaseContents",
    "Comment",
    "Contents",
    "CommentThread",
    "Subscriptions",
    "Users",
    "MODEL_INDICES",
]

MODEL_INDICES: tuple[type[BaseContents], ...] = (CommentThread, Comment)



"""
Handlers for the forum app.
"""

import logging
from typing import Any
from django.db.models.signals import post_save, post_delete
from django.dispatch import receiver

from forum.search import get_document_search_backend
from forum.utils import get_str_value_from_collection
from forum.models import Comment, CommentThread

log = logging.getLogger(__name__)


def handle_comment_thread_deletion(sender: Any, **kwargs: dict[str, str]) -> None:
    """
    Handle the deletion of a comment thread from the Elasticsearch index.

    Args:
        sender (Any): The model class that sends the signal.
        **kwargs (dict[str, Any]): Additional arguments, including 'comment_thread_id'.
    """
    thread_id = get_str_value_from_collection(kwargs, "comment_thread_id")
    search_backend = get_document_search_backend()
    search_backend.delete_document(sender.index_name, thread_id)


def handle_comment_deletion(sender: Any, **kwargs: dict[str, Any]) -> None:
    """
    Handle the deletion of a comment from the Elasticsearch index.

    Args:
        sender (Any): The model class that sends the signal.
        **kwargs (dict[str, Any]): Additional arguments, including 'comment_id'.
    """
    comment_id = get_str_value_from_collection(kwargs, "comment_id")
    search_backend = get_document_search_backend()
    search_backend.delete_document(sender.index_name, comment_id)


def handle_comment_thread_insertion(sender: Any, **kwargs: dict[str, Any]) -> None:
    """
    Handle the insertion of a comment thread into the Elasticsearch index.

    Args:
        sender (Any): The model class that sends the signal.
        **kwargs (dict[str, Any]): Additional arguments, including 'comment_thread_id'.
    """
    thread_id = get_str_value_from_collection(kwargs, "comment_thread_id")
    thread = sender().get(_id=thread_id)
    doc = sender().doc_to_hash(thread)
    get_document_search_backend().index_document(sender.index_name, thread_id, doc)
    log.info(f"Thread {thread_id} added to Elasticsearch index")


def handle_comment_insertion(sender: Any, **kwargs: dict[str, Any]) -> None:
    """
    Handle the insertion of a comment into the Elasticsearch index.

    Args:
        sender (Any): The model class that sends the signal.
        **kwargs (dict[str, Any]): Additional arguments, including 'comment_id'.
    """
    comment_id = get_str_value_from_collection(kwargs, "comment_id")
    comment = sender().get(_id=comment_id)
    doc = sender().doc_to_hash(comment)
    get_document_search_backend().index_document(sender.index_name, comment_id, doc)
    log.info(f"Comment {comment_id} added to Elasticsearch index")


def handle_comment_thread_updated(sender: Any, **kwargs: dict[str, Any]) -> None:
    """
    Handle the update of a comment thread in the Elasticsearch index.

    Args:
        sender (Any): The model class that sends the signal.
        **kwargs (dict[str, Any]): Additional arguments, including 'comment_thread_id'.
    """
    thread_id = get_str_value_from_collection(kwargs, "comment_thread_id")
    thread = sender().get(_id=thread_id)
    doc = sender().doc_to_hash(thread)
    get_document_search_backend().update_document(sender.index_name, thread_id, doc)
    log.info(f"Thread {thread_id} added to Elasticsearch index")


def handle_comment_updated(sender: Any, **kwargs: dict[str, Any]) -> None:
    """
    Handle the update of a comment in the Elasticsearch index.

    Args:
        sender (Any): The model class that sends the signal.
        **kwargs (dict[str, Any]): Additional arguments, including 'comment_id'.
    """
    comment_id = get_str_value_from_collection(kwargs, "comment_id")
    comment = sender().get(_id=comment_id)
    doc = sender().doc_to_hash(comment)
    get_document_search_backend().update_document(sender.index_name, comment_id, doc)
    log.info(f"Comment {comment_id} added to Elasticsearch index")


@receiver(post_delete, sender=CommentThread)
@receiver(post_delete, sender=Comment)
def handle_deletion(sender: Any, instance: Any, **kwargs: dict[str, Any]) -> None:
    """
    Handle the deletion of a comment thread or comment from the MySQL database.

    Args:
        sender (Any): The model class that sends the signal.
        instance (Any): The instance of the deleted comment thread or comment.
    """
    document_id = instance.id
    search_backend = get_document_search_backend()
    search_backend.delete_document(sender.index_name, document_id)
    log.info(f"{sender.__name__} {document_id} deleted from the search backend")


@receiver(post_save, sender=CommentThread)
@receiver(post_save, sender=Comment)
def handle_comment_thread_and_comment(
    sender: Any, instance: Any, created: bool, **kwargs: dict[str, Any]
) -> None:
    """
    Handle the insertion or update of a comment thread or comment in the MySQL database.

    Args:
        sender (Any): The model class that sends the signal.
        instance (Any): The instance of the comment thread or comment.
        created (bool): Indicates if the instance was created.
    """
    document_id = instance.id
    search_backend = get_document_search_backend()
    doc = instance.doc_to_hash()

    if created:
        search_backend.index_document(sender.index_name, document_id, doc)
        log.info(f"{sender.__name__} {document_id} added to the search backend")
    else:
        search_backend.update_document(sender.index_name, document_id, doc)
        log.info(f"{sender.__name__} {document_id} updated in the search backend")



"""Forum Comments API Views."""

from rest_framework import status
from rest_framework.permissions import AllowAny
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView

from forum.api.commentables import get_commentables_stats


class CommentablesCountAPIView(APIView):
    """
    API View to handle GET request for commentables count.
    """

    permission_classes = (AllowAny,)

    def get(self, request: Request, course_id: str) -> Response:
        """
        Retrieves a the threads count based on thread_type.

        Parameters:
            request (Request): The incoming request.
            course_id: The ID of the course.
        Body:
            Empty.
        Response:
            The threads count for the given course_id based on thread_type.
        """
        commentable_counts = get_commentables_stats(course_id)
        return Response(
            commentable_counts,
            status=status.HTTP_200_OK,
        )



"""
API for votes.
"""

from typing import Any, Optional

from forum.backend import get_backend
from forum.serializers.comment import CommentSerializer
from forum.serializers.thread import ThreadSerializer
from forum.serializers.votes import VotesInputSerializer
from forum.utils import ForumV2RequestError


def _get_thread_and_user(
    thread_id: str,
    user_id: str,
    course_id: Optional[str] = None,
) -> tuple[dict[str, Any], dict[str, Any]]:
    """
    Fetches the thread and user based on provided IDs.

    Args:
        thread_id (str): The ID of the thread.
        user_id (str): The ID of the user.

    Returns:
        tuple: The thread and user objects.

    Raises:
        ValueError: If the thread or user is not found.
    """
    backend = get_backend(course_id)()
    thread = backend.get_thread(thread_id)
    if not thread:
        raise ValueError("Thread not found")

    user = backend.get_user(user_id)
    if not user:
        raise ValueError("User not found")

    return thread, user


def _prepare_thread_response(
    thread: dict[str, Any], user: dict[str, Any], backend: Any
) -> dict[str, Any]:
    """
    Prepares the serialized response data after voting.

    Args:
        thread (dict): The thread data.
        user (dict): The user data.

    Returns:
        dict: The serialized response data.

    Raises:
        ValueError: If serialization fails.
    """
    context = {
        "id": str(thread["_id"]),
        **thread,
        "user_id": user["_id"],
        "username": user["username"],
        "type": "thread",
    }
    serializer = ThreadSerializer(data=context, backend=backend)
    if not serializer.is_valid():
        raise ValueError(serializer.errors)
    return serializer.data


def update_thread_votes(
    thread_id: str, user_id: str, value: str, course_id: Optional[str] = None
) -> dict[str, Any]:
    """
    Updates the votes for a thread.

    Args:
        thread_id (str): The ID of the thread.
        user_id (str): The ID of the user.
        value (str): The vote value ("up" or "down").
    """
    backend = get_backend(course_id)()
    data = {"user_id": user_id, "value": value}
    vote_serializer = VotesInputSerializer(data=data)

    if not vote_serializer.is_valid():
        raise ForumV2RequestError(vote_serializer.errors)

    try:
        thread, user = _get_thread_and_user(thread_id, user_id, course_id=course_id)
    except ValueError as error:
        raise ForumV2RequestError(str(error)) from error

    if vote_serializer.data["value"] == "up":
        is_updated = backend.upvote_content(
            thread_id, user_id, entity_type="CommentThread"
        )
    else:
        is_updated = backend.downvote_content(
            thread_id, user_id, entity_type="CommentThread"
        )

    if is_updated:
        thread = backend.get_thread(thread_id) or {}

    return _prepare_thread_response(thread, user, backend)


def delete_thread_vote(
    thread_id: str, user_id: str, course_id: Optional[str] = None
) -> dict[str, Any]:
    """
    Deletes the vote for a thread.

    Args:
        thread_id (str): The ID of the thread.
        user_id (str): The ID of the user.
    """
    backend = get_backend(course_id)()
    try:
        _, user = _get_thread_and_user(thread_id, user_id, course_id=course_id)
    except ValueError as error:
        raise ForumV2RequestError(str(error)) from error

    deleted_thread = None
    if backend.remove_vote(thread_id, user_id, entity_type="CommentThread"):
        deleted_thread = backend.get_thread(thread_id)

    if not deleted_thread:
        raise ForumV2RequestError("Thread not found")

    return _prepare_thread_response(deleted_thread, user, backend)


def _get_comment_and_user(
    comment_id: str, user_id: str, backend: Any
) -> tuple[dict[str, Any], dict[str, Any]]:
    """
    Fetches the comment and user based on provided IDs.

    Args:
        comment_id (str): The ID of the comment.
        user_id (str): The ID of the user.

    Returns:
        tuple: The comment and user objects.

    Raises:
        ValueError: If the comment or user is not found.
    """
    comment = backend.get_comment(comment_id)
    if not comment:
        raise ValueError("Comment not found")

    user = backend.get_user(user_id)
    if not user:
        raise ValueError("User not found")

    return comment, user


def _prepare_comment_response(
    comment: dict[str, Any], user: dict[str, Any], backend: Any
) -> dict[str, Any]:
    """
    Prepares the serialized response data after voting.

    Args:
        comment (dict): The comment data.
        user (dict): The user data.

    Returns:
        dict: The serialized response data.

    Raises:
        ValueError: If serialization fails.
    """
    context = {
        "id": str(comment["_id"]),
        **comment,
        "user_id": user["_id"],
        "username": user["username"],
        "type": "comment",
        "thread_id": str(comment.get("comment_thread_id", None)),
    }
    serializer = CommentSerializer(data=context, backend=backend)
    if not serializer.is_valid():
        raise ValueError(serializer.errors)
    return serializer.data


def update_comment_votes(
    comment_id: str, user_id: str, value: str, course_id: Optional[str] = None
) -> dict[str, Any]:
    """
    Updates the votes for a comment.

    Args:
        comment_id (str): The ID of the comment.
        user_id (str): The ID of the user.
        value (str): The vote value ("up" or "down").
    """
    backend = get_backend(course_id)()
    data = {"user_id": user_id, "value": value}
    vote_serializer = VotesInputSerializer(data=data)

    if not vote_serializer.is_valid():
        raise ForumV2RequestError(vote_serializer.errors)

    try:
        _, user = _get_comment_and_user(comment_id, user_id, backend)
    except ValueError as error:
        raise ForumV2RequestError(str(error)) from error

    if vote_serializer.data["value"] == "up":
        is_updated = backend.upvote_content(comment_id, user_id, entity_type="Comment")
    else:
        is_updated = backend.downvote_content(
            comment_id, user_id, entity_type="Comment"
        )

    updated_comment = None
    if is_updated:
        updated_comment = backend.get_comment(comment_id)

    if not updated_comment:
        raise ForumV2RequestError("Comment not found")

    return _prepare_comment_response(updated_comment, user, backend)


def delete_comment_vote(
    comment_id: str, user_id: str, course_id: Optional[str] = None
) -> dict[str, Any]:
    """
    Deletes the vote for a comment.

    Args:
        comment_id (str): The ID of the comment.
        user_id (str): The ID of the user.
    """
    backend = get_backend(course_id)()
    try:
        _, user = _get_comment_and_user(comment_id, user_id, backend)
    except ValueError as error:
        raise ForumV2RequestError(str(error)) from error

    deleted_comment = None
    if backend.remove_vote(comment_id, user_id, entity_type="Comment"):
        deleted_comment = backend.get_comment(comment_id)

    if not deleted_comment:
        raise ForumV2RequestError("Comment not found")

    return _prepare_comment_response(deleted_comment, user, backend)



"""
Native Python Commenttables APIs.
"""

from forum.backend import get_backend


def get_commentables_stats(course_id: str) -> dict[str, int]:
    """
    Get the threads count based on thread_type and group them by commentable_id.

    Parameters:
        course_id: The ID of the course.
    Body:
        Empty.
    Response:
        The threads count for the given course_id based on thread_type.
        e.g.
        reponse = {'course': {'discussion': 1, 'question': 1}}
    """
    backend = get_backend(course_id)()
    return backend.get_commentables_counts_based_on_type(course_id)
